/* eslint-disable no-magic-numbers */
import {
  Arrow,
  ArrowType,
  LayerLine,
  LayerOval,
  LayerRect,
  LayerText,
  LineType,
} from '@/types/models';
import { MapInstance } from '@/types/map';
import { UsePointToProjectionResult } from '@/utils/map/usePointToProjection';
import { Feature } from 'ol';
import { LineString, MultiPolygon, Point } from 'ol/geom';
import Text from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/text/Text';
import Polygon from 'ol/geom/Polygon';
import VectorSource from 'ol/source/Vector';
import VectorLayer from 'ol/layer/Vector';
import getEllipseCoords from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/coords/getEllipseCoords';
import getStyle from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/style/getStyle';
import {
  HorizontalAlign,
  VerticalAlign,
} from '@/components/Map/MapViewer/MapViewerVector/MapViewerVector.types';
import getMultiPolygon from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/elements/getMultiPolygon';
import Style from 'ol/style/Style';
import { BLACK_COLOR } from '@/components/Map/MapViewer/MapViewerVector/MapViewerVector.constants';
import getRotation from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/coords/getRotation';

export interface LayerProps {
  texts: Array<LayerText>;
  rects: Array<LayerRect>;
  ovals: Array<LayerOval>;
  lines: Array<LayerLine>;
  visible: boolean;
  layerId: string;
  lineTypes: Array<LineType>;
  arrowTypes: Array<ArrowType>;
  mapInstance: MapInstance;
  pointToProjection: UsePointToProjectionResult;
}

export default class Layer {
  texts: Array<LayerText>;

  rects: Array<LayerRect>;

  ovals: Array<LayerOval>;

  lines: Array<LayerLine>;

  lineTypes: Array<LineType>;

  arrowTypes: Array<ArrowType>;

  textFeatures: Array<Feature<Point>>;

  rectFeatures: Array<Feature<Polygon>>;

  ovalFeatures: Array<Feature<Polygon>>;

  lineFeatures: Array<Feature<LineString>>;

  arrowFeatures: Array<Feature<MultiPolygon>>;

  vectorSource: VectorSource<
    Feature<Point> | Feature<Polygon> | Feature<LineString> | Feature<MultiPolygon>
  >;

  vectorLayer: VectorLayer<
    VectorSource<Feature<Point> | Feature<Polygon> | Feature<LineString> | Feature<MultiPolygon>>
  >;

  constructor({
    texts,
    rects,
    ovals,
    lines,
    visible,
    layerId,
    lineTypes,
    arrowTypes,
    mapInstance,
    pointToProjection,
  }: LayerProps) {
    this.texts = texts;
    this.rects = rects;
    this.ovals = ovals;
    this.lines = lines;
    this.lineTypes = lineTypes;
    this.arrowTypes = arrowTypes;
    this.textFeatures = this.getTextsFeatures(mapInstance, pointToProjection);
    this.rectFeatures = this.getRectsFeatures(pointToProjection);
    this.ovalFeatures = this.getOvalsFeatures(pointToProjection);
    const { linesFeatures, arrowsFeatures } = this.getLinesFeatures(pointToProjection);
    this.lineFeatures = linesFeatures;
    this.arrowFeatures = arrowsFeatures;
    this.vectorSource = new VectorSource({
      features: [
        ...this.textFeatures,
        ...this.rectFeatures,
        ...this.ovalFeatures,
        ...this.lineFeatures,
        ...this.arrowFeatures,
      ],
    });
    this.vectorLayer = new VectorLayer({
      source: this.vectorSource,
      visible,
    });
    this.vectorLayer.set('id', layerId);
  }

  private getTextsFeatures = (
    mapInstance: MapInstance,
    pointToProjection: UsePointToProjectionResult,
  ): Array<Feature<Point>> => {
    const textObjects = this.texts.map(text => {
      return new Text({
        x: text.x,
        y: text.y,
        zIndex: text.z,
        width: text.width,
        height: text.height,
        fontColor: text.color,
        fontSize: text.fontSize,
        text: text.notes,
        verticalAlign: text.verticalAlign as VerticalAlign,
        horizontalAlign: text.horizontalAlign as HorizontalAlign,
        pointToProjection,
        mapInstance,
      });
    });
    return textObjects.map(text => text.feature);
  };

  private getRectsFeatures = (
    pointToProjection: UsePointToProjectionResult,
  ): Array<Feature<Polygon>> => {
    return this.rects.map(rect => {
      const polygon = new Polygon([
        [
          pointToProjection({ x: rect.x, y: rect.y }),
          pointToProjection({ x: rect.x + rect.width, y: rect.y }),
          pointToProjection({ x: rect.x + rect.width, y: rect.y + rect.height }),
          pointToProjection({ x: rect.x, y: rect.y + rect.height }),
        ],
      ]);
      const polygonStyle = getStyle({
        geometry: polygon,
        borderColor: rect.borderColor,
        fillColor: rect.fillColor,
        lineWidth: rect.lineWidth,
        zIndex: rect.z,
      });
      const rectFeature = new Feature<Polygon>({
        geometry: polygon,
      });
      rectFeature.setStyle(polygonStyle);
      return rectFeature;
    });
  };

  private getOvalsFeatures = (
    pointToProjection: UsePointToProjectionResult,
  ): Array<Feature<Polygon>> => {
    return this.ovals.map(oval => {
      const coords = getEllipseCoords({
        x: oval.x + oval.width / 2,
        y: oval.y + oval.height / 2,
        height: oval.height,
        width: oval.width,
        pointToProjection,
        points: 36,
      });
      const polygon = new Polygon([coords]);
      const polygonStyle = getStyle({
        geometry: polygon,
        borderColor: oval.borderColor,
        fillColor: { rgb: 0, alpha: 0 },
        lineWidth: oval.lineWidth,
        zIndex: oval.z,
      });
      const ovalFeature = new Feature<Polygon>({
        geometry: polygon,
      });
      ovalFeature.setStyle(polygonStyle);
      return ovalFeature;
    });
  };

  private getLinesFeatures = (
    pointToProjection: UsePointToProjectionResult,
  ): {
    linesFeatures: Array<Feature<LineString>>;
    arrowsFeatures: Array<Feature<MultiPolygon>>;
  } => {
    const linesFeatures: Array<Feature<LineString>> = [];
    const arrowsFeatures: Array<Feature<MultiPolygon>> = [];

    this.lines.forEach(line => {
      const points = line.segments
        .map((segment, index) => {
          if (index === 0) {
            return [
              pointToProjection({ x: segment.x1, y: segment.y1 }),
              pointToProjection({ x: segment.x2, y: segment.y2 }),
            ];
          }
          return [pointToProjection({ x: segment.x2, y: segment.y2 })];
        })
        .flat();

      const firstSegment = line.segments[0];
      const startArrowRotation = getRotation(
        [firstSegment.x1, firstSegment.y1],
        [firstSegment.x2, firstSegment.y2],
      );
      const shortenedX1 = firstSegment.x1 + line.startArrow.length * Math.cos(startArrowRotation);
      const shortenedY1 = firstSegment.y1 - line.startArrow.length * Math.sin(startArrowRotation);
      points[0] = pointToProjection({ x: shortenedX1, y: shortenedY1 });

      const lastSegment = line.segments[line.segments.length - 1];
      const endArrowRotation = getRotation(
        [lastSegment.x1, lastSegment.y1],
        [lastSegment.x2, lastSegment.y2],
      );
      const shortenedX2 = lastSegment.x2 - line.endArrow.length * Math.cos(endArrowRotation);
      const shortenedY2 = lastSegment.y2 - line.endArrow.length * Math.sin(endArrowRotation);
      points[points.length - 1] = pointToProjection({ x: shortenedX2, y: shortenedY2 });

      const lineString = new LineString(points);

      let lineDash;
      const lineTypeFound = this.lineTypes.find(type => type.name === line.lineType);
      if (lineTypeFound) {
        lineDash = lineTypeFound.pattern;
      }
      const lineStyle = getStyle({
        geometry: lineString,
        borderColor: line.color,
        lineWidth: line.width,
        lineDash,
        zIndex: line.z,
      });
      const lineFeature = new Feature<LineString>({
        geometry: lineString,
      });
      lineFeature.setStyle(lineStyle);
      linesFeatures.push(lineFeature);

      arrowsFeatures.push(
        ...this.getLineArrowsFeatures({
          line,
          pointToProjection,
          startArrowX: firstSegment.x1,
          startArrowY: firstSegment.y1,
          startArrowRotation,
          endArrowX: shortenedX2,
          endArrowY: shortenedY2,
          endArrowRotation,
        }),
      );
    });
    return { linesFeatures, arrowsFeatures };
  };

  private getLineArrowsFeatures = ({
    line,
    pointToProjection,
    startArrowX,
    startArrowY,
    startArrowRotation,
    endArrowX,
    endArrowY,
    endArrowRotation,
  }: {
    line: LayerLine;
    pointToProjection: UsePointToProjectionResult;
    startArrowX: number;
    startArrowY: number;
    startArrowRotation: number;
    endArrowX: number;
    endArrowY: number;
    endArrowRotation: number;
  }): Array<Feature<MultiPolygon>> => {
    const arrowsFeatures: Array<Feature<MultiPolygon>> = [];
    const startArrowFeature = this.getLineArrowFeature(
      line.startArrow,
      startArrowX,
      startArrowY,
      line.z,
      startArrowRotation,
      line.width,
      pointToProjection,
    );
    if (startArrowFeature) {
      arrowsFeatures.push(startArrowFeature);
    }

    const endArrowFeature = this.getLineArrowFeature(
      line.endArrow,
      endArrowX,
      endArrowY,
      line.z,
      endArrowRotation,
      line.width,
      pointToProjection,
    );
    if (endArrowFeature) {
      arrowsFeatures.push(endArrowFeature);
    }
    return arrowsFeatures;
  };

  private getLineArrowFeature = (
    arrow: Arrow,
    x: number,
    y: number,
    zIndex: number,
    rotation: number,
    lineWidth: number,
    pointToProjection: UsePointToProjectionResult,
  ): undefined | Feature<MultiPolygon> => {
    const arrowShapes = this.arrowTypes.find(arrowType => arrowType.arrowType === arrow.arrowType)
      ?.shapes;
    if (!arrowShapes) {
      return undefined;
    }
    const arrowMultiPolygon = getMultiPolygon({
      x,
      y: y - arrow.length / 2,
      width: arrow.length,
      height: arrow.length,
      shapes: arrowShapes,
      pointToProjection,
    });
    const arrowStyles: Array<Style> = [];
    arrowMultiPolygon.forEach(polygon => {
      const style = getStyle({
        geometry: polygon,
        zIndex,
        borderColor: BLACK_COLOR,
        fillColor: BLACK_COLOR,
        lineWidth,
      });
      arrowStyles.push(style);
      polygon.rotate(rotation, pointToProjection({ x, y }));
    });
    const arrowFeature = new Feature({
      geometry: new MultiPolygon(arrowMultiPolygon),
    });
    arrowFeature.setStyle(arrowStyles);
    return arrowFeature;
  };
}
